function template_fit, jd_i, mag_i, mag_i_err, jd_v, $
                       mag_v, mag_v_err, short_struc=short_struc, $
                       long_struc=long_struc, fit_err=fit_err , $
                       plots=plots, ns=ns, red_chi=red_chi, $
                       p_guess=p_guess, yfit=yfit, norm_err=norm_err
;+
; NAME:  template_fit
;
;
;
; PURPOSE: fit magnitudes and periods of input cephied light curves
;
;
;
; CATEGORY: curve fitting
;
;
;
; CALLING SEQUENCE:
;
;
;
; INPUTS:
;        
;jd_i:  I-band observing times in Julian day
;mag_i: I-band magnitudes
;mag_i_err:  The errors of the i-band magnitudes
;jd_v:  V-band observing times in Julian day
;mag_v: V-band magnitudes
;mag_v_err:  V-band magnitude errors
;short_struc:  structure containing the Cepheid template information
;             to be used for short (P < 10 day) stars
;long_struc:  what you think
;
;
;
;
;
;
;
; OPTIONAL INPUTS:
;
;ns: 0-4.  Number of additional shape parameters to fit.  Defualt 0.
;p_guess:  Initial guess parameters.  By default it uses the median
;         magnitudes and a series of initial period guesses.
;
; KEYWORD PARAMETERS:
;
;
;
; OUTPUTS:
;result:  The best fit parameters in the form [m_i, m_v, Period,
;        phase, break, pca1,pca2,pca3,pca4]
;
;m_i,m_v = average I and V magnitudes for the model
;Period:  best fit period (days)
;phase:  the phase it happened to shift to get a good fit (days, I think)
;break:  The index where the data switches from being I band to V band.
;
;
;
; OPTIONAL OUTPUTS:
;pca1-4:  if ns is set, these are returned
;fit_err:  The errors in result as calculated from the covariance
;         matrix
;norm_err:  The error normalized so chi^2=1.  Early Monte Carlo
;          testing shows norm error is a good estimate of the actual
; error for well sampled light curves.  It can actually _over_
; estimate the error in cases where the sampling is low.  It can also
; be radically small if the sampling does not cover a whole period.  
;yfit:  The best fit points
;red_chi:  The reduced chi^2 value (hopefully close to 1, but whatever)
;
;
;
; COMMON BLOCKS:
;
;
;
; SIDE EFFECTS:
;nausia and vomitting.
;
;
; RESTRICTIONS:
;
;
;
; PROCEDURE:
;
;
;
; EXAMPLE:
;
;
;
; MODIFICATION HISTORY:
;finalized 11/07, PY
;-





if not keyword_set(ns) then ns=0 ;number of extra shape params

;make an early simple guess at what the period is
if not keyword_set(p_guess) then begin
  p1=period_simp(jd_i,mag_i, range=[1.,30.],/harms)
  p2=period_simp(jd_v,mag_v, range=[1.,30.],/harms)
  p_guess=[p1,p2,findgen(30)+1.5]               ;hell try them both!
  p_guess=[p_guess,(findgen(20)+1)*2]
endif


jd=[jd_i,jd_v]
mag=[mag_i,mag_v]
mag_err=[mag_i_err,mag_v_err]

ni=n_elements(jd_i)


p0=[median(mag_i,/double), median(mag_v), 0., 0., ni]

if ns ne 0 then p0=[p0,fltarr(ns)]

parinfo=replicate({fixed:0,limited:[0,0],limits:[0.D,0],mpminstep:0.D},5+ns)
parinfo([4]).fixed=1 ;fix the break index
parinfo(3).limited=[1,0]
parinfo(3).limits=[0,40]

ptry=p_guess 

phase_try=[.3, .6] ;fractional phase-shifts to try

np=n_elements(ptry)
nph=n_elements(phase_try)

all_fits=fltarr(np,nph,5+ns)
all_bn=fltarr(np,nph)
for i=0,np-1 do begin
  p0(2)=ptry(i)
  if p0(2) lt 10. then ceph_struc=short_struc else $
                       ceph_struc=long_struc
;  parinfo(3).limits=[0., p0(2)]
  for j=0,nph-1 do begin
    p0(3)=p0(2)*phase_try(j)
    temp_fit=mpfitfun('gen_lc', jd,mag,mag_err, p0, parinfo=parinfo, $
                      bestnorm=bestnorm, $
                      functargs={struct:ceph_struc}, perror=perror,/quiet)
    ;stop
    all_fits(i,j,*)=temp_fit
    all_bn(i,j)=bestnorm
    if i eq 0 and j eq 0 then begin
      bn=bestnorm
      best_fit=temp_fit
    endif

    if bestnorm lt bn and bestnorm ne 1.0 then begin ;the fit gives a
                                ;lower chi-sq, I think it gives 1.0 if
                                ;there's an error
      bn=bestnorm
      best_fit=temp_fit
    endif
  endfor
endfor

;run one final one to get all the info

if best_fit(2) lt 10. then ceph_struc=short_struc else ceph_struc=long_struc
parinfo(3).limits=[0,p0(2)]


best_fit=mpfitfun('gen_lc', jd,mag,mag_err, best_fit, parinfo=parinfo, $
                      bestnorm=bestnorm, perror=perror, dof=dof,yfit=yfit,$
                      functargs={struct:ceph_struc},/quiet)


red_chi=bestnorm/dof

fit_err=perror;*sqrt(bestnorm/dof)
norm_err=perror*sqrt(bestnorm/dof)

if keyword_set(plots) then begin
  plotsym, 0,/fill
  x=fold(jd, best_fit(2))
  mplot, x, mag, /nodata, xtitle='Phase', ytitle='magnitudes'
  oploterror, x,mag,mag_err, psym=8
  oplot_split, x,yfit, best_fit(4)
endif


;stop




return, best_fit
end

